@@ -1,11 +1,14 @@
 import os
 import subprocess
+import sys
 import sysconfig
-from distutils import log
-from distutils.command.clean import clean
-from typing import List, Set, Tuple, Type, cast
+import logging
+
+from typing import List, Optional, Set, Tuple, Type, TypeVar, cast
+from functools import partial
 
 from setuptools.command.build_ext import build_ext
+
 from setuptools.command.install import install
 from setuptools.command.install_lib import install_lib
 from setuptools.command.install_scripts import install_scripts
@@ -13,13 +16,24 @@ from setuptools.command.sdist import sdist
 from setuptools.dist import Distribution
 from typing_extensions import Literal
 
-from .extension import RustBin, RustExtension
+from .build import _get_bdist_wheel_cmd
+from .extension import Binding, RustBin, RustExtension, Strip
 
 try:
     from wheel.bdist_wheel import bdist_wheel
 except ImportError:
     bdist_wheel = None
 
+if sys.version_info[:2] >= (3, 11):
+    from tomllib import load as toml_load
+else:
+    from tomli import load as toml_load
+
+
+logger = logging.getLogger(__name__)
+
+T = TypeVar("T", bound=RustExtension)
+
 
 def add_rust_extension(dist: Distribution) -> None:
     sdist_base_class = cast(Type[sdist], dist.cmdclass.get("sdist", sdist))
@@ -145,29 +159,33 @@ def add_rust_extension(dist: Distribution) -> None:
 
         def run(self) -> None:
             if self.distribution.rust_extensions:
-                log.info("running build_rust")
+                logger.info("running build_rust")
                 build_rust = self.get_finalized_command("build_rust")
                 build_rust.inplace = self.inplace
                 build_rust.target = self.target
                 build_rust.verbose = self.verbose
-                options = self.distribution.get_cmdline_options().get("bdist_wheel", {})
-                plat_name = options.get("plat-name") or self.plat_name
-                build_rust.plat_name = plat_name
+                build_rust.plat_name = self._get_wheel_plat_name() or self.plat_name
                 build_rust.run()
 
             build_ext_base_class.run(self)
 
+        def _get_wheel_plat_name(self) -> Optional[str]:
+            cmd = _get_bdist_wheel_cmd(self.distribution)
+            return cast(Optional[str], getattr(cmd, "plat_name", None))
+
     dist.cmdclass["build_ext"] = build_ext_rust_extension
 
-    clean_base_class = cast(Type[clean], dist.cmdclass.get("clean", clean))
+    clean_base_class = dist.cmdclass.get("clean")
 
-    class clean_rust_extension(clean_base_class):  # type: ignore[misc,valid-type]
-        def run(self) -> None:
-            clean_base_class.run(self)
-            if not self.dry_run:
-                self.run_command("clean_rust")
+    if clean_base_class is not None:
 
-    dist.cmdclass["clean"] = clean_rust_extension
+        class clean_rust_extension(clean_base_class):  # type: ignore[misc,valid-type]
+            def run(self) -> None:
+                super().run()
+                if not self.dry_run:
+                    self.run_command("clean_rust")
+
+        dist.cmdclass["clean"] = clean_rust_extension
 
     install_base_class = cast(Type[install], dist.cmdclass.get("install", install))
 
@@ -282,6 +300,33 @@ def rust_extensions(
         add_rust_extension(dist)
 
 
+def pyprojecttoml_config(dist: Distribution) -> None:
+    try:
+        with open("pyproject.toml", "rb") as f:
+            cfg = toml_load(f).get("tool", {}).get("setuptools-rust")
+    except FileNotFoundError:
+        return None
+
+    if cfg:
+        modules = map(partial(_create, RustExtension), cfg.get("ext-modules", []))
+        binaries = map(partial(_create, RustBin), cfg.get("bins", []))
+        dist.rust_extensions = [*modules, *binaries]  # type: ignore[attr-defined]
+        rust_extensions(dist, "rust_extensions", dist.rust_extensions)  # type: ignore[attr-defined]
+
+
+def _create(constructor: Type[T], config: dict) -> T:
+    kwargs = {
+        # PEP 517/621 convention: pyproject.toml uses dashes
+        k.replace("-", "_"): v
+        for k, v in config.items()
+    }
+    if "binding" in config:
+        kwargs["binding"] = Binding[config["binding"]]
+    if "strip" in config:
+        kwargs["strip"] = Strip[config["strip"]]
+    return constructor(**kwargs)
+
+
 _CARGO_VENDOR_CONFIG = b"""
 [source.crates-io]
 replace-with = "vendored-sources"
